#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "huffman.h"

void exit_if(int condition, const char *prefix) {
    if (condition) {
        perror(prefix);
        exit(EXIT_FAILURE);
    }
}

static const char *special_chars = "\\\"{|}<>";

static void pp_letter(FILE *out, int letter)
{
    const char *pos = strchr(special_chars, letter);
    if (pos != NULL && *pos) // remove the end of string
        fprintf(out, "\\%c", letter);
    else if (letter == EOS)
        fprintf(out, "EOS");
    else if (letter == ' ')
        fprintf(out, "space");
    else if (letter < 32 || letter >= 127)
        fprintf(out, "0x%.x", letter);
    else
        fprintf(out, "%c", letter);
}

static void pp_labels(FILE *out, struct tree *node)
{
    if (is_leaf(node)) {
        fprintf(out, "\t\"%p\" [shape=record, label=\"{", node);
        pp_letter(out, node->letter);
        fprintf(out, "|%d}\"];\n", node->value);
    } else {
        fprintf(out, "\t\"%p\" [label=\"%d\"];\n", node, node->value);
        if (node->left)
            pp_labels(out, node->left);
        if (node->right)
            pp_labels(out, node->right);
    }
}
static void pp_edges(FILE *out, struct tree *node)
{ // FIXME
    if (is_leaf(node)) return;
    if (node->left)
        fprintf(out, "\t\"%p\" -> \"%p\" [label=0];\n", node, node->left);
    if (node->right)
        fprintf(out, "\t\"%p\" -> \"%p\" [label=1];\n", node, node->right);
    if (node->left)
        pp_edges(out, node->left);
    if (node->right)
        pp_edges(out, node->right);
}

void print_tree_to_dot(FILE *out, struct tree *root)
{
    fprintf(out, "digraph G {\n");
    fprintf(out, "\tgraph [ranksep=0];\n");
    if (root) {
        pp_labels(out, root);
        pp_edges(out, root);
    }
    fprintf(out, "};\n");
}

void print_bits(FILE *out, unsigned long long x, int bits)
{
    while (bits --) {
        fprintf(out, "%.1d", (int)(x >> bits) & 1);
    }
}

void print_compress_table(FILE *out, const struct bit_buffer table[])
{
    for (int i = 0; i < MAX_UCHAR; i++)
        if (table[i].count > 0) {
            pp_letter(out, i);
            fprintf(out, ":\t");
            print_bits(out, table[i].bits, table[i].count);
            fputc('\n', out);
        }
}
